import asyncio
import logging
from typing import List, Optional, Union

import a_sync
from brownie import chain
from multicall import Call

from y import convert
from y.classes.common import ERC20, ContractBase
from y.contracts import Contract
from y.datatypes import AddressOrContract, AnyAddressType, Block, UsdPrice
from y.exceptions import ContractNotVerified
from y.networks import Network
from y.utils.logging import _get_price_logger
from y.utils.multicall import fetch_multicall
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


v1_pools = {
    Network.Mainnet: [
        "0x398eC7346DcD622eDc5ae82352F02bE94C62d119"
    ],
}.get(chain.id, [])

v2_pools = {
    Network.Mainnet: [
        "0x7d2768dE32b0b80b7a3454c06BdAc94A69DDc7A9", # aave
        "0x7937D4799803FbBe595ed57278Bc4cA21f3bFfCB", # aave amm
        "0xcE744a9BAf573167B2CF138114BA32ed7De274Fa", # umee
    ],
    Network.Polygon: [
        "0x8dFf5E27EA6b7AC08EbFdf9eB090F32ee9a30fcf", # aave
    ],
    Network.Avalanche: [
        "0x70BbE4A294878a14CB3CDD9315f5EB490e346163", # blizz
    ],
}.get(chain.id, [])        

v3_pools = {
    Network.Mainnet: [
        "0x87870Bca3F3fD6335C3F4ce8392D69350B4fA4E2", # aave v3
    ],
    Network.Optimism: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD", # aave v3
    ],
    Network.Arbitrum: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD", # aave v3
    ],
    Network.Harmony: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD", # aave v3
    ],
    Network.Arbitrum: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD", # aave v3
    ],
    Network.Fantom: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD", # aave v3
    ],
    Network.Avalanche: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD", # aave v3
    ],
    Network.Polygon: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD", # aave v3
    ],
}.get(chain.id, [])

class AaveMarketBase(ContractBase):
    def __contains__(self, __o: object) -> bool:
        if not self.asynchronous:
            cls = self.__class__.__name__
            raise RuntimeError(f"'self.asynchronous' must be False to use {cls}.__contains__.\nYou may wish to use {cls}.is_atoken instead.")
        return convert.to_address(__o) in self.atokens
    
    async def contains(self, __o: object) -> bool:
        contains = convert.to_address(__o) in await self.__atokens__(sync=False)
        logger.debug(f'{self} contains {__o}: {contains}')
        return contains


class AaveMarketV1(AaveMarketBase):
    @a_sync.a_sync(ram_cache_maxsize=256)
    async def underlying(self, token_address: AddressOrContract) -> ERC20:
        underlying = await raw_call(token_address, 'underlyingAssetAddress()',output='address', sync=False)
        return ERC20(underlying)
    
    @a_sync.aka.cached_property
    async def atokens(self) -> List[ERC20]:
        reserves_data = await Call(self.address, ['getReserves()(address[])'], [[self.address,None]]).coroutine()
        reserves_data = reserves_data[self.address]
        reserves_data = fetch_multicall(*[[self.contract, 'getReserveData', reserve] for reserve in reserves_data])
        atokens = [ERC20(reserve['aTokenAddress'], asynchronous=self.asynchronous) for reserve in reserves_data]
        logger.info(f'loaded {len(atokens)} v1 atokens for {self.__repr__()}')
        return atokens


class AaveMarketV2(AaveMarketBase):
    @a_sync.a_sync(ram_cache_maxsize=256)
    async def underlying(self, token_address: AddressOrContract) -> ERC20:
        underlying = await raw_call(token_address, 'UNDERLYING_ASSET_ADDRESS()',output='address', sync=False)
        logger.debug(f"underlying: {underlying}")
        return ERC20(underlying, asynchronous=self.asynchronous)

    @a_sync.aka.cached_property
    async def atokens(self) -> List[ERC20]:
        reserves = await Call(self.address, ['getReservesList()(address[])'], [[self.address,None]]).coroutine()
        reserves = reserves[self.address]
        reserves_data = await asyncio.gather(*[
            Call(
                self.address,
                ['getReserveData(address)((uint256,uint128,uint128,uint128,uint128,uint128,uint40,address,address,address,address,uint8))',reserve],
            ).coroutine()
            for reserve in reserves
        ])

        try:
            atokens = [ERC20(reserve_data[7], asynchronous=self.asynchronous) for reserve_data in reserves_data]
            logger.info(f'loaded {len(atokens)} v2 atokens for {self.__repr__()}')
            return atokens
        except TypeError as e: # TODO figure out what to do about non verified aave markets
            logger.exception(e)
            logger.warning(f'failed to load tokens for {self.__repr__()}')
            return []


class AaveMarketV3(AaveMarketBase):
    @a_sync.a_sync(ram_cache_maxsize=256)
    async def underlying(self, token_address: AddressOrContract) -> ERC20:
        underlying = await raw_call(token_address, 'UNDERLYING_ASSET_ADDRESS()',output='address', sync=False)
        logger.debug(f"underlying: {underlying}")
        return ERC20(underlying, asynchronous=self.asynchronous)

    @a_sync.aka.cached_property
    async def atokens(self) -> List[ERC20]:
        reserves = await Call(self.address, ['getReservesList()(address[])'], [[self.address,None]]).coroutine()
        reserves = reserves[self.address]
        reserves_data = await asyncio.gather(*[self.contract.getReserveData.coroutine(reserve) for reserve in reserves])

        try:
            atokens = [ERC20(reserve_data[8], asynchronous=self.asynchronous) for reserve_data in reserves_data]
            logger.info(f'loaded {len(atokens)} v3 atokens for {self.__repr__()}')
            return atokens
        except TypeError as e: # TODO figure out what to do about non verified aave markets
            logger.exception(e)
            logger.warning(f'failed to load tokens for {self.__repr__()}')
            return []


class AaveRegistry(a_sync.ASyncGenericSingleton):
    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous

    @a_sync.aka.cached_property
    async def pools(self) -> List[Union[AaveMarketV1, AaveMarketV2]]:
        v1, v2, v3 = await asyncio.gather(
            self.__pools_v1__(sync=False),
            self.__pools_v2__(sync=False), 
            self.__pools_v3__(sync=False),
        )
        return v1 + v2 + v3
    
    @a_sync.aka.cached_property
    async def pools_v1(self) -> List[AaveMarketV1]:
        pools = [AaveMarketV1(pool, asynchronous=self.asynchronous) for pool in v1_pools]
        logger.debug(f"{self} v1 pools {pools}")
        return pools
    
    @a_sync.aka.cached_property
    async def pools_v2(self) -> List[AaveMarketV2]:
        pools = [AaveMarketV2(pool, asynchronous=self.asynchronous) for pool in v2_pools]
        logger.debug(f"{self} v2 pools {pools}")
        return pools
    
    @a_sync.aka.cached_property
    async def pools_v3(self) -> List[AaveMarketV2]:
        pools = [AaveMarketV3(pool, asynchronous=self.asynchronous) for pool in v3_pools]
        logger.debug(f"{self} v3 pools {pools}")
        return pools
    
    async def pool_for_atoken(self, token_address: AnyAddressType) -> Optional[Union[AaveMarketV1, AaveMarketV2]]:
        pools = await self.__pools__(sync=False)
        for pool in pools:
            if await pool.contains(token_address, sync=False):
                return pool

    def __contains__(self, __o: object) -> bool:
        if self.asynchronous:
            raise RuntimeError(f"'self.asynchronous' must be False to use AaveRegistry.__contains__.\nYou may wish to use AaveRegistry.is_atoken instead.")
        return any(__o in pool for pool in self.pools)

    @a_sync.a_sync(cache_type='memory')
    async def is_atoken(self, token_address: AnyAddressType) -> bool:
        logger = _get_price_logger(token_address, block=None)
        is_atoken = any(await asyncio.gather(*[pool.contains(token_address, sync=False) for pool in await self.__pools__(sync=False)]))
        logger.debug(f"is_atoken: {is_atoken}")
        return is_atoken
    
    async def is_wrapped_atoken_v2(self, token_address: AnyAddressType) -> bool:
        # NOTE: Not sure if this wrapped version is actually related to aave but this works for pricing purposes.
        try:
            contract = await Contract.coroutine(token_address)
            attrs = "ATOKEN", "STATIC_ATOKEN_LM_REVISION", "staticToDynamicAmount"
            return all(hasattr(contract, attr) for attr in attrs)
        except ContractNotVerified:
            return False
        
    async def is_wrapped_atoken_v3(self, token_address: AnyAddressType) -> bool:
        # NOTE: Not sure if this wrapped version is actually related to aave but this works for pricing purposes.
        try:
            contract = await Contract.coroutine(token_address)
            attrs = "ATOKEN", "AAVE_POOL", "UNDERLYING"
            return all(hasattr(contract, attr) for attr in attrs)
        except ContractNotVerified:
            return False
    
    @a_sync.a_sync(cache_type='memory')
    async def underlying(self, token_address: AddressOrContract) -> ERC20:
        pool: Union[AaveMarketV1, AaveMarketV2] = await self.pool_for_atoken(token_address, sync=False)
        return await pool.underlying(token_address, sync=False)
    
    async def get_price(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        underlying = await self.underlying(token_address, sync=False)
        return await underlying.price(block, sync=False)

    async def get_price_wrapped_v2(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        return await self._get_price_wrapped(token_address, 'staticToDynamicAmount', block=block)

    async def get_price_wrapped_v3(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        return await self._get_price_wrapped(token_address, 'convertToAssets', block=block)
    
    async def _get_price_wrapped(self, token_address: AddressOrContract, method: str, block: Optional[Block] = None) -> UsdPrice:
        contract, scale = await asyncio.gather(Contract.coroutine(token_address), ERC20(token_address, asynchronous=True).scale)
        underlying, price_per_share = await asyncio.gather(
            contract.ATOKEN.coroutine(block_identifier=block),  # NOTE: We can probably cache this without breaking anything
            getattr(contract, method).coroutine(scale, block_identifier=block),
        )
        price_per_share /= scale
        return price_per_share * await ERC20(underlying, asynchronous=True).price(block)


aave = AaveRegistry(asynchronous=True)



import asyncio
import logging
from collections import Counter, defaultdict
from itertools import zip_longest
from typing import Any, AsyncGenerator, Dict, List, Optional

import a_sync
import eth_retry
from a_sync.primitives import ThreadsafeSemaphore
from brownie import web3
from brownie.convert.datatypes import EthAddress
from brownie.network.event import EventDict, _decode_logs
from eth_typing import ChecksumAddress
from toolz import groupby
from web3.middleware.filter import block_ranges
from web3.types import LogReceipt

from y.constants import thread_pool_executor
from y.contracts import contract_creation_block_async
from y.datatypes import Address, Block
from y.utils.cache import memory
from y.utils.dank_mids import dank_w3
from y.utils.middleware import BATCH_SIZE

logger = logging.getLogger(__name__)


def decode_logs(logs: List[LogReceipt]) -> EventDict:
    """
    Decode logs to events and enrich them with additional info.
    """
    decoded = _decode_logs(logs)
    for i, log in enumerate(logs):
        setattr(decoded[i], "block_number", log["blockNumber"])
        setattr(decoded[i], "transaction_hash", log["transactionHash"])
        setattr(decoded[i], "log_index", log["logIndex"])
    return decoded


@a_sync.a_sync(default='sync')
async def get_logs_asap(
    address: Optional[Address],
    topics: Optional[List[str]],
    from_block: Optional[Block] = None,
    to_block: Optional[Block] = None,
    verbose: int = 0
) -> List[Any]:

    if from_block is None:
        from_block = 0 if address is None else await contract_creation_block_async(address, True)
    if to_block is None:
        to_block = await dank_w3.eth.block_number

    ranges = list(block_ranges(from_block, to_block, BATCH_SIZE))
    if verbose > 0:
        logger.info('fetching %d batches', len(ranges))
    
    batches = await asyncio.gather(*[_get_logs_async(address, topics, start, end) for start, end in ranges])
    return [log for batch in batches for log in batch]


async def get_logs_asap_generator(
    address: Optional[Address],
    topics: Optional[List[str]] = None,
    from_block: Optional[Block] = None,
    to_block: Optional[Block] = None,
    chronological: bool = True,
    run_forever: bool = False,
    run_forever_interval: int = 60,
    verbose: int = 0
) -> AsyncGenerator[List[LogReceipt], None]:
    # NOTE: If you don't need the logs in order, you will get your logs faster if you set `chronological` to False.

    if from_block is None:
        from_block = 0 if address is None else await contract_creation_block_async(address, True)
    if to_block is None:
        to_block = await dank_w3.eth.block_number
    elif run_forever:
        raise TypeError(f'`to_block` must be None if `run_forever` is True.')
    if from_block > to_block:
        raise ValueError(f"from_block must be <= to_block. You passed from_block: {from_block} to_block: {to_block}.")
    while True:
        ranges = list(block_ranges(from_block, to_block, BATCH_SIZE))
        if verbose > 0:
            logger.info('fetching %d batches', len(ranges))
        coros = [_get_logs_async(address, topics, start, end) for start, end in ranges]
        if chronological:
            for logs in await asyncio.gather(*coros):
                logs = yield logs
        else:
            for logs in asyncio.as_completed(coros, timeout=None):
                yield await logs
        if not run_forever:
            return
        
        await asyncio.sleep(run_forever_interval)
        
        # Find start and end block for next loop
        current_block = await dank_w3.eth.block_number
        while current_block <= to_block:
            await asyncio.sleep(run_forever_interval)
            current_block = await dank_w3.eth.block_number
        from_block = to_block + 1 if to_block + 1 <= current_block else current_block
        to_block = current_block


def logs_to_balance_checkpoints(logs) -> Dict[EthAddress,int]:
    """
    Convert Transfer logs to `{address: {from_block: balance}}` checkpoints.
    """
    balances = Counter()
    checkpoints = defaultdict(dict)
    for block, block_logs in groupby('blockNumber', logs).items():
        events = decode_logs(block_logs)
        for log in events:
            # ZERO_ADDRESS tracks -totalSupply
            sender, receiver, amount = log.values()  # there can be several different aliases
            balances[sender] -= amount
            checkpoints[sender][block] = balances[sender]
            balances[receiver] += amount
            checkpoints[receiver][block] = balances[receiver]
    return checkpoints


def checkpoints_to_weight(checkpoints, start_block: Block, end_block: Block) -> float:
    total = 0
    for a, b in zip_longest(list(checkpoints), list(checkpoints)[1:]):
        if a < start_block or a > end_block:
            continue
        b = min(b, end_block) if b else end_block
        total += checkpoints[a] * (b - a) / (end_block - start_block)
    return total

@a_sync.a_sync(executor=thread_pool_executor)
def _get_logs(
    address: Optional[ChecksumAddress],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    if end - start == BATCH_SIZE - 1:
        response = _get_logs_batch_cached(address, topics, start, end)
    else:
        response = _get_logs_no_cache(address, topics, start, end)
    for log in response:
        if address and log.address != address:
            ''' I have this due to a corrupt cache on my local box that I would prefer not to lose. '''
            ''' It will not impact your scripts. ''' 
            response.remove(log)
    return response

get_logs_semaphores = defaultdict(lambda: ThreadsafeSemaphore(16))

def _get_semaphore_key(address, topics) -> tuple:
    if isinstance(address, list):
        address.sort()
    addr_key = tuple(address) if isinstance(address, list) else address
    topics_key = tuple(tuple(t) if isinstance(t, list) else t for t in topics) if isinstance(topics, list) else topics
    return addr_key, topics_key

async def _get_logs_async(address, topics, start, end) -> List[LogReceipt]:
    async with get_logs_semaphores[_get_semaphore_key(address, topics)]:
        return await _get_logs(address, topics, start, end, asynchronous=True)

@eth_retry.auto_retry
def _get_logs_no_cache(
    address: Optional[ChecksumAddress],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    logger.debug(f'fetching logs {start} to {end}')
    try:
        if address is None:
            response = web3.eth.get_logs({"topics": topics, "fromBlock": start, "toBlock": end})
        elif topics is None:
            response = web3.eth.get_logs({"address": address, "fromBlock": start, "toBlock": end})
        else:
            response = web3.eth.get_logs({"address": address, "topics": topics, "fromBlock": start, "toBlock": end})
    except Exception as e:
        errs = [
            "Service Unavailable for url:",
            "exceed maximum block range",
            "block range is too wide",
        ]
        if any(err in str(e) for err in errs):
            logger.debug('your node is having trouble, breaking batch in half')
            batch_size = (end - start + 1)
            half_of_batch = batch_size // 2
            batch1_end = start + half_of_batch
            batch2_start = batch1_end + 1
            batch1 = _get_logs_no_cache(address, topics, start, batch1_end)
            batch2 = _get_logs_no_cache(address, topics, batch2_start, end)
            response = batch1 + batch2
        else:
            raise
    return response



##############################
# Leaving this here so as to not upset people's caches
##############################


@memory.cache()
def _get_logs_batch_cached(
    address: Optional[str],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    return _get_logs_no_cache(address, topics, start, end)




import asyncio
import logging
from typing import Optional

import a_sync
from brownie import ZERO_ADDRESS, chain

from y.constants import usdc
from y.contracts import Contract
from y.datatypes import Address, Block, UsdPrice
from y.exceptions import UnsupportedNetwork, continue_if_call_reverted
from y.networks import Network
from y.utils.raw_calls import _decimals

logger = logging.getLogger(__name__)

V1 = {
    Network.Mainnet: "0xc0a47dFe034B400B47bDaD5FecDa2621de6c4d95",
}.get(chain.id,None)


class UniswapV1(a_sync.ASyncGenericBase):
    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
        super().__init__()
    
    @a_sync.aka.cached_property
    async def factory(self) -> Contract:
        if V1 is None:
            raise UnsupportedNetwork(f"UniswapV1 does not suppport chainid {chain.id}")
        return await Contract.coroutine(V1)            
        
    @a_sync.a_sync(ram_cache_maxsize=256)
    async def get_exchange(self, token_address: Address) -> Optional[Contract]:
        factory = await self.__factory__(sync=False)
        exchange = await factory.getExchange.coroutine(token_address)
        if exchange == ZERO_ADDRESS:
            return None
        return await Contract.coroutine(exchange)

    async def get_price(self, token_address: Address, block: Optional[Block]) -> Optional[UsdPrice]:
        exchange, usdc_exchange, decimals = await asyncio.gather(
            self.get_exchange(token_address, sync=False),
            self.get_exchange(usdc, sync=False),
            _decimals(token_address, block, sync=False),
        )
        if exchange is None:
            return None
        
        try:
            eth_bought = await exchange.getTokenToEthInputPrice.coroutine(10 ** decimals, block_identifier=block)
            usdc_bought = await usdc_exchange.getEthToTokenInputPrice.coroutine(eth_bought, block_identifier=block) / 1e6
            fees = 0.997 ** 2
            return UsdPrice(usdc_bought / fees)
        except ValueError as e:
            if 'invalid jump destination' in str(e):
                return None
            continue_if_call_reverted(e)




import pytest

from tests.fixtures import blocks_for_contract
from y.prices.lending.compound import CToken, compound

CTOKENS = [ctoken.address for troller in compound.trollers.values() for ctoken in troller.__markets__(sync=True)]

@pytest.mark.parametrize('token', CTOKENS)
def test_compound_pricing_sync(token):
    print(token)
    #if convert.to_address(token) == '0x892B14321a4FCba80669aE30Bd0cd99a7ECF6aC0':
    #    continue  # creth is broken
    ctoken = CToken(token)
    for block in blocks_for_contract(token):
        print(f'underlying per ctoken = {ctoken.underlying_per_ctoken(block)}')
        price = ctoken.get_price(block)
        assert price, 'Failed to fetch price.'
        print(f'                price = {price}')

@pytest.mark.parametrize('token', CTOKENS)
async def test_compound_pricing_async(token):
    print(token)
    #if convert.to_address(token) == '0x892B14321a4FCba80669aE30Bd0cd99a7ECF6aC0':
    #    continue  # creth is broken
    ctoken = CToken(token, asynchronous=True)
    for block in blocks_for_contract(token):
        print(f'underlying per ctoken = {await ctoken.underlying_per_ctoken(block)}')
        price = await compound.get_price(token, block)
        assert price, 'Failed to fetch price.'
        print(f'                price = {price}')
